/*
 * $Id: ApplyQuery.java,v 1.4 2003/01/29 15:59:16 triolet Exp $
 *
 * Transmorpher
 *
 * Copyright (C) 2001-2002 Fluxmedia and INRIA Rhne-Alpes.
 * Copyright (C) INRIA Rhne-Alpes, 2003.
 *
 * http://www.fluxmedia.fr - http://transmorpher.inrialpes.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package fr.fluxmedia.transmorpher.graph;
import fr.fluxmedia.transmorpher.engine.TProcess;
import fr.fluxmedia.transmorpher.engine.TProcessComposite;

import fr.fluxmedia.transmorpher.utils.Parameters;
import fr.fluxmedia.transmorpher.utils.Writer;

import java.io.IOException;

/**
	*The ApplyQuery Class introduces the use of a query in a process body. Its structure is the following:
	* <br/><code><i>
	*	&lt;apply-query type="type" ref="name" id="id" in="channel"&gt;<br/>
	* {&lt;with-param&gt;}<br/>
	*	&lt;/apply-query&gt;<br/>
	* </i></code><br/>
 *@author    Jerome.Euzenat@inrialpes.fr
 *@since     jdk 1.3 / SAX 2.0
 */
public class ApplyQuery extends ApplyImpl {

	/**
	 *Creates a simple instance of ApplyQuery
	 */
	public ApplyQuery() {
		this((String)null, (String)null);
	}

	/**
	 *Creates an instance of ApplyQuery with an id and a type
	 *
	 *@param  id  	The id of the instance to create
	 *@param  type  The type of the instance to create
	 */
	public ApplyQuery(String id, String type) {
		this(id, type, null);
	}

	/**
	 *Creates an instance of ApplyQuery with an id ,a type and a process
	 *
	 *@param  id  		The id of the instance to create
	 *@param  type  	The type of the instance to create
	 *@param  process	The process that owns this instance
	 */
	public ApplyQuery(String id, String type, Process process) {
		this(id, type, process, null);
	}

	/**
	 *Creates an instance of ApplyQuery with an id ,a type, a process and a reference
	 *
	 *@param  id  			The id of the instance to create
	 *@param  type  		The type of the instance to create
	 *@param  process		The process that owns this instance
	 *@param  reference The name of the Query called by this instance
	 */
	public ApplyQuery(String id, String type, Process process, String reference) {
		super(id, type, process,reference, 1 /*NbIn  */, 1 /*NbOut  */);
	}

	/**
	 * Propagate the nullify of its sole output to all the input.
	 *
	 *@param  out  Description of the Parameter
	 *@param  in   Description of the Parameter
	 */
	public void retroNull(PortList out, PortList in) {
		nullified = out.getPort(0).getChannel().nullifiedP();
		in.getPort(0).getChannel().setVisited();
		if (nullified) {
			in.getPort(0).getChannel().nullify();
		}
		visited = true;

		update(null);
	}

	/**
	 *  Prints an XML description of this ApplyQuery instance
	 */
	public void generateXML() {
		System.out.print("      <apply-query id=\"" + id + "\" type=\"" + getType() + "\" ref=\"" + getRef() + "\" ");
		System.out.print("in=\"");
		inPorts.generateXML();
		System.out.print("\" out=\"");
		outPorts.generateXML();
		if (parameters.isEmpty()) {
			System.out.println("\"/>");
		} else {
			System.out.println("\">");
			parameters.generateXMLCalls();
			System.out.println("      </apply-query>");
		}
	}

	/**
	 *  Creates a TApplyQuery corresponding to this ApplyQuery object and adds it to the composite
	 * 	process in the execution structure.
	 *
	 *@param  currentProcess  the container for this serializer
	 */
	public final void createProcess(TProcessComposite currentProcess) {//throws TransformerException, TransformerConfigurationException, SAXException, IOException

		String[] vPortIn = inPorts().toStringList();
		String[] vPortOut = outPorts().toStringList();
		String className = this.getProcess().getTransmorpher().getClassForType(getType());

		// get the name corresponding to the generated stylesheet and pass it in parameter
		String ref = getAttributes().getStringParameter("ref");
		Transmorpher tr = getProcess().getTransmorpher();
		String f = tr.findTransformation(ref).getAttributes().getStringParameter("file");
		// static because generated by Transmorpher from the query name
		// NOTE: this is specific from tmq and not the query constructor...
		getParameters().setParameter("file", f);
		Object[] params = {(Object)vPortIn, (Object)vPortOut, (Object)getParameters(), (Object)getAttributes()};
		TProcess vCurrent = this.newProcess(className, params);
		currentProcess.addProcess(vCurrent);

		update(vCurrent);
	}//end proc

	/**
	 *  
	 */
	public void setUp() {
		if (getProcess().getTransmorpher().findTransformation(getRef()) == null) {
			System.err.println("warning : unknown query : " + getRef());
		}//end if

		update(null);

	}//end proc

	/**
	 *  Prints a java code description of this ApplyQuery instance
	 *
	 *@param  file             A writer used for printing in afile
	 *@exception  IOException  when an IO errors occur
	 */
	public void generateJavaCode(Writer file) throws IOException {
		file.writeln("");
		file.writeln(4, "//------------------Generation of ApplyQuery: " + getId() + " -------------------");
		generatePorts(file);
		// Fetch the file to be processed by XSLT (this is very xsltish)
		String ref = getAttributes().getStringParameter("ref");
		Transmorpher tr = getProcess().getTransmorpher();
		String f = tr.findTransformation(ref).getAttributes().getStringParameter("file");
		// Added just for being generated afterwards
		getParameters().setParameter("file", f);
		generateParameters(file);
		String className = this.getProcess().getTransmorpher().getClassForType(getType());
		file.writeln(4, "tmCurrentProcess = new " + className + "(portIn,portOut,tmParameters,tmAttributes);");
		file.writeln(4, "tmCurrentProcess.setName(\"" + getId() + "\");");
		file.writeln("");
		file.writeln(4, "tmCurrentCompositeProcess.addProcess(tmCurrentProcess);");
	}//end proc

}

